// -*- Mode: Go; indent-tabs-mode: t -*-

/*
 * Copyright (C) 2015 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package cloud

import (
	"fmt"
	"os"
	"strings"
	"testing"

	check "gopkg.in/check.v1"
	"launchpad.net/snappy-tests-job/testhelpers"
	"launchpad.net/snappy-tests-job/utils"
)

const (
	defaultRelease = "1504"
	defaultChannel = "edge"
	imageListTpl   = `| e37ef727-014d-48e8-b0e4-94cf39e1dbbc | ubuntu-released/ubuntu-wily-alpha1-amd64-server-20150624-disk1.img                                   | ACTIVE |                                      |
| 633e3e02-bf9e-4de4-ba2b-361603a88233 | ubuntu-released/ubuntu-wily-alpha1-i386-server-20150624-disk1.img                                    | ACTIVE |                                      |
| 566e3e02-bf9e-4de4-ba2b-361603a88df6 | %s                            | ACTIVE |               `
	imageListMultipleTpl = `| 633e3e02-bf9e-4de4-ba2b-361603a88233 | %s                                    | ACTIVE |                                      |
| 566e3e02-bf9e-4de4-ba2b-361603a88df6 | %s                            | ACTIVE |              |
| 633e3e02-bf9e-4de4-ba2b-361603a88233 | %s                                    | ACTIVE |                                      |`

	launchOutputTpl = `+--------------------------------------+-----------------------------------------------------------------------------------------------------------+
| Property                             | Value                                                                                                     |
+--------------------------------------+-----------------------------------------------------------------------------------------------------------+
| OS-DCF:diskConfig                    | MANUAL                                                                                                    |
| OS-EXT-AZ:availability_zone          | nova                                                                                                      |
| OS-EXT-STS:power_state               | 0                                                                                                         |
| OS-EXT-STS:task_state                | scheduling                                                                                                |
| OS-EXT-STS:vm_state                  | building                                                                                                  |
| OS-SRV-USG:launched_at               | -                                                                                                         |
| OS-SRV-USG:terminated_at             | -                                                                                                         |
| accessIPv4                           |                                                                                                           |
| accessIPv6                           |                                                                                                           |
| adminPass                            | HB5gEKXzp8Ls                                                                                              |
| config_drive                         |                                                                                                           |
| created                              | 2015-07-27T09:28:49Z                                                                                      |
| flavor                               | m1.small (2)                                                                                              |
| hostId                               |                                                                                                           |
| id                                   | %s                                                                                                        |
| image                                | ubuntu-released/ubuntu-wily-alpha1-amd64-server-20150624-disk1.img (e37ef727-014d-48e8-b0e4-94cf39e1dbbc) |
| key_name                             | ues-snappy-integration-tests_lcy02                                                                        |
| metadata                             | {}                                                                                                        |
| name                                 | www                                                                                                       |
| os-extended-volumes:volumes_attached | []                                                                                                        |
| progress                             | 0                                                                                                         |
| security_groups                      | ssh                                                                                                       |
| status                               | BUILD                                                                                                     |
| tenant_id                            | 245eb8a420d9466d919eb7ee3a54cc23                                                                          |
| updated                              | 2015-07-27T09:28:49Z                                                                                      |
| user_id                              | 0534127f70534300baf48e8ac1134734                                                                          |
+--------------------------------------+-----------------------------------------------------------------------------------------------------------+`
	describeOutputTpl = `
+--------------------------------------+-----------------------------------------------------------------------------------------------------------+
| Property                             | Value                                                                                                     |
+--------------------------------------+-----------------------------------------------------------------------------------------------------------+
| OS-DCF:diskConfig                    | MANUAL                                                                                                    |
| OS-EXT-AZ:availability_zone          | nova                                                                                                      |
| OS-EXT-STS:power_state               | 1                                                                                                         |
| OS-EXT-STS:task_state                | -                                                                                                         |
| OS-EXT-STS:vm_state                  | active                                                                                                    |
| OS-SRV-USG:launched_at               | 2015-07-27T09:29:27.000000                                                                                |
| OS-SRV-USG:terminated_at             | -                                                                                                         |
| accessIPv4                           |                                                                                                           |
| accessIPv6                           |                                                                                                           |
| generic_networkid_string network     | %s                                                                                                        |
| config_drive                         |                                                                                                           |
| created                              | 2015-07-27T09:28:49Z                                                                                      |
| flavor                               | m1.small (2)                                                                                              |
| hostId                               | f9b4d12d39c28f1f0ec970622c3c04348ab019b4b9571fe5567d76b6                                                  |
| id                                   | 4e4aff22-2e5e-42b5-b179-38ed9e6fb0e6                                                                      |
| image                                | ubuntu-released/ubuntu-wily-alpha1-amd64-server-20150624-disk1.img (e37ef727-014d-48e8-b0e4-94cf39e1dbbc) |
| key_name                             | ues-snappy-integration-tests_lcy02                                                                        |
| metadata                             | {}                                                                                                        |
| name                                 | www                                                                                                       |
| os-extended-volumes:volumes_attached | []                                                                                                        |
| progress                             | 0                                                                                                         |
| security_groups                      | ssh                                                                                                       |
| status                               | ACTIVE                                                                                                    |
| tenant_id                            | 245eb8a420d9466d919eb7ee3a54cc23                                                                          |
| updated                              | 2015-07-27T09:29:27Z                                                                                      |
| user_id                              | 0534127f70534300baf48e8ac1134734                                                                          |
+--------------------------------------+-----------------------------------------------------------------------------------------------------------+`
	instanceIP   = "1.1.1.1"
	novaUserName = "my_nova_user_name"
	osRegionName = "my_os_region_name"
)

var _ = check.Suite(&canonistackSuite{})

type canonistackSuite struct {
	fakeUtil               *testhelpers.FakeUtils
	subject                Clouder
	instanceID             string
	imageID                string
	flavor                 string
	novaUserNameBack       string
	osRegionNameBack       string
	imageList              string
	describeOutputPending  string
	describeOutputFinished string
}

func Test(t *testing.T) { check.TestingT(t) }

func (s *canonistackSuite) SetUpTest(c *check.C) {
	s.fakeUtil = testhelpers.NewFakeUtils()
	s.flavor = "m1.quitelarge"

	s.subject = NewCanonistackClient(s.flavor, s.fakeUtil)

	s.instanceID = "instanceID1"
	s.imageID = imgTemplate(defaultRelease, defaultChannel)

	s.novaUserNameBack = os.Getenv("NOVA_USERNAME")
	s.osRegionNameBack = os.Getenv("OS_REGION_NAME")

	os.Setenv("NOVA_USERNAME", novaUserName)
	os.Setenv("OS_REGION_NAME", osRegionName)

	s.imageList = fmt.Sprintf(imageListTpl, s.imageID)
	s.describeOutputPending = fmt.Sprintf(describeOutputTpl, "")
	s.describeOutputFinished = fmt.Sprintf(describeOutputTpl, instanceIP)

	s.fakeUtil.ExecReturnValues = nil
	s.fakeUtil.ExecReturnValues = append(s.fakeUtil.ExecReturnValues, s.imageList)
	launchOutput := fmt.Sprintf(launchOutputTpl, s.instanceID)
	s.fakeUtil.ExecReturnValues = append(s.fakeUtil.ExecReturnValues, launchOutput)
	s.fakeUtil.ExecReturnValues = append(s.fakeUtil.ExecReturnValues, s.describeOutputFinished)
}

func (s *canonistackSuite) TearDownTest(c *check.C) {
	os.Setenv("NOVA_USERNAME", s.novaUserNameBack)
	os.Setenv("OS_REGION_NAME", s.osRegionNameBack)
}

func (s *canonistackSuite) TestCreateInstanceCallsListImages(c *check.C) {
	s.subject.CreateInstance(defaultRelease, defaultChannel)

	cmdParams := utils.NewExecCommandParams("", strings.Fields(listImagesCmd), false)

	c.Assert(s.fakeUtil.ExecCalls[cmdParams.String()], check.Equals, 1,
		check.Commentf("Expected euca list images called once"))
}

func (s *canonistackSuite) TestCreateInstanceReturnsErrorOnListError(c *check.C) {
	s.fakeUtil.FailExec = true
	_, err := s.subject.CreateInstance(defaultRelease, defaultChannel)

	c.Assert(err, check.NotNil, check.Commentf("Expected error from euca list images"))
}

func (s *canonistackSuite) TestCreateInstanceDoesNotCreateNewInstanceOnListError(c *check.C) {
	s.fakeUtil.FailExec = true
	s.subject.CreateInstance(defaultRelease, defaultChannel)

	cmd := s.launchInstanceCmd(s.imageID, s.flavor)
	cmdParams := utils.NewExecCommandParams("", strings.Fields(cmd), false)

	c.Assert(s.fakeUtil.ExecCalls[cmdParams.String()], check.Equals, 0,
		check.Commentf("Called LaunchInstance on ListImages error"))
}

func (s *canonistackSuite) TestCreateInstanceCallsLaunchInstance(c *check.C) {
	s.subject.CreateInstance(defaultRelease, defaultChannel)

	cmd := s.launchInstanceCmd(s.imageID, s.flavor)
	cmdParams := utils.NewExecCommandParams("", strings.Fields(cmd), false)

	c.Assert(s.fakeUtil.ExecCalls[cmdParams.String()], check.Equals, 1,
		check.Commentf("Expected call to LaunchInstance not occurred"))
}

func (s *canonistackSuite) TestCreateInstanceReturnsInstanceIP(c *check.C) {
	ip, _ := s.subject.CreateInstance(defaultRelease, defaultChannel)

	c.Assert(ip, check.Equals, instanceIP,
		check.Commentf("Expected instance IP not returned %s, got %s", instanceIP, ip))
}

func (s *canonistackSuite) TestKillInstanceCallsKillCommand(c *check.C) {
	s.subject.CreateInstance(defaultRelease, defaultChannel)
	s.subject.KillInstance()

	cmd := s.killInstanceCmd()
	cmdParams := utils.NewExecCommandParams("", strings.Fields(cmd), false)

	c.Assert(s.fakeUtil.ExecCalls[cmdParams.String()], check.Equals, 1,
		check.Commentf("Expected call to KillInstance not occurred"))
}

func (s *canonistackSuite) TestKillInstanceNotCalledIfMisisngInstance(c *check.C) {
	back := s.instanceID
	s.instanceID = ""
	defer func() { s.instanceID = back }()

	s.subject.KillInstance()

	cmd := s.killInstanceCmd()
	cmdParams := utils.NewExecCommandParams("", strings.Fields(cmd), false)

	c.Assert(s.fakeUtil.ExecCalls[cmdParams.String()], check.Equals, 0,
		check.Commentf("Call to KillInstance without active instance"))
}

func (s *canonistackSuite) TestCreateInstanceCallsLaunchInstanceWithLatestImage(c *check.C) {
	s.fakeUtil.ExecReturnValues = nil

	imageIDNew := s.imageID + "-" + "20150109"
	imageIDOld := s.imageID + "-" + "20150107"
	imageIDOlder := s.imageID + "-" + "20150105"

	imageList := fmt.Sprintf(imageListMultipleTpl,
		imageIDOlder, imageIDNew, imageIDOld)
	s.fakeUtil.ExecReturnValues = append(s.fakeUtil.ExecReturnValues, imageList)
	launchOutput := fmt.Sprintf(launchOutputTpl, imageIDNew)
	s.fakeUtil.ExecReturnValues = append(s.fakeUtil.ExecReturnValues, launchOutput)
	s.fakeUtil.ExecReturnValues = append(s.fakeUtil.ExecReturnValues, s.describeOutputFinished)

	s.subject.CreateInstance(defaultRelease, defaultChannel)

	cmd := s.launchInstanceCmd(imageIDNew, s.flavor)
	cmdParams := utils.NewExecCommandParams("", strings.Fields(cmd), false)

	c.Assert(s.fakeUtil.ExecCalls[cmdParams.String()], check.Equals, 1,
		check.Commentf("Expected call to LaunchInstance with newest image not occurred"))
}

func imgTemplate(release, channel string) string {
	return fmt.Sprintf(imgTpl, release, channel)
}

func (s *canonistackSuite) launchInstanceCmd(imgID, flavor string) string {
	return fmt.Sprintf(launchInstanceCmdTpl,
		novaUserName, osRegionName, flavor, imgID, os.Getpid())
}

func (s *canonistackSuite) killInstanceCmd() string {
	return fmt.Sprintf(killInstanceCmdTpl, s.instanceID)
}
