// -*- Mode: Go; indent-tabs-mode: t -*-

/*
 * Copyright (C) 2015 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package testhelpers

import (
	"errors"

	"launchpad.net/snappy-tests-job/utils"
)

// FakeUtils is a fake implementation of the Utilizer interface for testing
type FakeUtils struct {
	ExecCalls        map[string]int
	MkDirCalls       map[string]int
	RemoveAllCalls   map[string]int
	FailExec         bool
	FailMkDir        bool
	ExecReturnValues []string
	ExecReturnIndex  int
}

// NewFakeUtils is the FakeUtils constructors, it initializes the map members
func NewFakeUtils() *FakeUtils {
	return &FakeUtils{
		ExecCalls:      make(map[string]int),
		MkDirCalls:     make(map[string]int),
		RemoveAllCalls: make(map[string]int)}
}

// ExecCommand implementation on FakeUtils
func (fk *FakeUtils) ExecCommand(params *utils.ExecCommandParams) (output string, err error) {
	fk.ExecCalls[params.String()]++
	if fk.FailExec {
		err = errors.New("ExecCommand failed!")
	}

	if returnValuesLen := len(fk.ExecReturnValues); returnValuesLen > 0 && fk.ExecReturnIndex <= returnValuesLen-1 {
		output = fk.ExecReturnValues[fk.ExecReturnIndex]
		fk.ExecReturnIndex++
	}
	return
}

// PrepareTargetDir implementation on FakeUtils
func (fk *FakeUtils) PrepareTargetDir(path string) error {
	fk.MkDirCalls[path]++
	if fk.FailMkDir {
		return errors.New("PrepareTargetDir failed!")
	}
	return nil
}

// RemoveAll implementation on FakeUtils
func (fk *FakeUtils) RemoveAll(path string) error {
	fk.RemoveAllCalls[path]++
	return nil
}
